import collections, sys

from . import debug, text, actions

class RefernceError(Exception):
  pass

class ParseError(Exception):
  def __init__(self, message, object, tokens=()):
    self.message = message
    self.object = object
    self.tokens = tokens

  def __str__(self):
    if len(self.tokens) <= 7:
      tks = repr(list(self.tokens))
    else:
      tks = repr(list(self.tokens[:8]))[:-1] + ' ...'
    return """
  {message}
While parsing:
  {object}
Within:
  {parent}
Subsequent tokens:
  {tokens}
""".format(
  message = self.message,
  object = self.object,
  parent = self.object.parent,
  tokens = tks,
)

def indent(thing):
  return '\n  '.join(str(thing).split('\n'))

def parse_TEXT(data):
  '''
  Parses a Temporal EXTension file and returns a TEXT object.
  '''
  AST = assemble_AST(data)
  # First pass, to populate the object:
  text = AST.object()
  # Second pass, to resolve references:
  text = AST.object((text.lookup, text.add))
  return text

def lookup_value_type(parent_type, field_name):
  '''
  Looks up the parent type and field name of a field to determine the type of
  ASTNode that corresponds to the value for that field. Uses the 
  '''
  vtypes = {}
  for typ in reversed(parent_type.__mro__):
    if typ in FIELD_VALUE_TYPES:
      vtypes.update(FIELD_VALUE_TYPES[typ])
  if field_name in vtypes:
    return vtypes[field_name]
  return Literal

class Reference:
  '''
  A reference to a value. Just has a single 'ref' property.
  '''
  ref = None
  def __init__(self, ref):
    self.ref = ref

  def __eq__(self, other):
    return self.ref == other.ref

  def __ne__(self, other):
    return not self == other

  def __hash__(self):
    return hash(self.ref) + 17

  def __str__(self):
    return "a reference to '{}'".format(self.ref)

  def resolve(self, lookup, _):
    obj = lookup(self.ref)
    if not obj:
      raise RefernceError(
        "Reference to '{}' failed to resolve.".format(self.ref)
      )
    else:
      return obj

class GenRef(Reference):
  typ = None

  def __init__(self, ref, typ):
    super().__init__(ref)
    self.typ = typ

  def __str__(self):
    return "a generative reference to '{}'".format(self.ref)

  def resolve(self, lookup, create):
    obj = lookup(self.ref)
    if not obj:
      try:
        obj = self.typ(name=self.ref)
      except:
        raise RefernceError(
          ("Generative reference of type '{}' to '{}' failed to generate " +\
           "referent.").format(self.typ, self.ref)
        )
    return obj


# The parent class for all node types:

class ASTNode:
  '''
  A node in the Abstract Syntax Tree for a TEXT. Each node type defines its own
  parse and object methods. The parse method takes a string of tokens, removes
  tokens from the front in order to fill in the internal data for its type, and
  returns the remaining tokens. This often involves constructing one or more
  other ASTNode objects and calling their parse methods. Each object method
  constructs and returns an object appropriate to the ASTNode. This often
  involves recursively calling the object method of child nodes.
  '''
  parent = None
  def __init__(self, parent):
    self.parent = parent

  def __str__(self):
    return "an AST node"

  def __repr__(self):
    return "<an abstract syntax tree node for a TEXT>"

  def parse_list(self, tokens, eachtype, allowcompact=True):
    '''
    Parses a list of eachtype elements and returns the result (as a list),
    along with the remaining tokens.
    '''
    items = []
    if tuple(tokens[:2]) == (':', '\n'):
      tokens = tokens[2:]
      while tokens[0] in ('', '\n'):
        tokens = tokens[1:]
      while tokens[0] != 'end':
        item = eachtype(self)
        tokens = item.parse(tokens)
        items.append(item)
        while tokens[0] in ('', '\n'):
          tokens = tokens[1:]
      if tokens and tokens[0] == 'end':
        tokens = tokens[1:]
      if tokens and tokens[0] == '\n':
        tokens = tokens[1:]
    elif allowcompact:
      while tokens[0] != '\n':
        item = eachtype(self)
        tokens = item.parse(tokens)
        items.append(item)
      tokens = tokens[1:]
    else:
      raise ParseError(
        "{}s must use the ': ... end' listing style.".format(
          self.__class__.__name__
        ),
        self,
        tokens
      )
    return items, tokens

#####################
# Base node classes #
#####################

class Type(ASTNode):
  name = None

  def __str__(self):
    return "Type '{}'".format(self.name)

  def parse(self, tokens):
    debug.debug("Parsing", "Type:start")
    debug.indent("Parsing")
    self.name = tokens[0]
    debug.exdent("Parsing")
    debug.debug("Parsing", "Type:end")
    return tokens[1:]

  def object(self, _=False):
    result = None
    if self.name in dir(text):
      result = getattr(text, self.name)
    elif self.name in dir(actions):
      result = getattr(actions, self.name)
    elif self.name in dir(__builtins__):
      result = getattr(__builtins__, self.name)
    elif self.name in globals():
      result = globals()[self.name]
    else:
      raise ParseError(
        "Couldn't resolve type '{}'!".format(self.name),
        self,
        ["<unknown>"],
      )
    debug.debug("Construction", "Resolved type '{}' as {}".format(
      self.name,
      result
    ))
    return result

class Literal(ASTNode):
  value = None

  def __str__(self):
    return "Literal '{}'".format(self.value)

  def parse(self, tokens):
    debug.debug("Parsing", "Literal:start")
    debug.indent("Parsing")
    self.value = tokens[0]
    debug.exdent("Parsing")
    debug.debug("Parsing", "Literal:end")
    return tokens[1:]

  def object(self, _=False):
    return self.value

class Number(ASTNode):
  value = None

  def __str__(self):
    return "Number '{}'".format(self.value)

  def parse(self, tokens):
    debug.debug("Parsing", "Number:start")
    debug.indent("Parsing")
    self.value = tokens[0]
    debug.exdent("Parsing")
    debug.debug("Parsing", "Number:end")
    return tokens[1:]

  def object(self, _=False):
    return float(self.value)

class ImpliedTrue(ASTNode):
  def __str__(self):
    return "ImpliedTrue"

  def parse(self, tokens):
    debug.debug("Parsing", "ImpliedTrue:start")
    debug.debug("Parsing", "ImpliedTrue:end")
    return tokens

  def object(self, _=False):
    return True

########################
# Complex node classes #
########################

class FieldMap(ASTNode):
  fields = set()

  def __str__(self):
    return "FieldMap with fields:\n  {}".format(
      indent(', '.join(str(f) for f in self.fields))
    )

  def parse(self, tokens):
    debug.debug("Parsing", "FieldMap:start")
    debug.indent("Parsing")
    items, tokens = self.parse_list(tokens, Field)
    self.fields = set(items)
    debug.exdent("Parsing")
    debug.debug("Parsing", "FieldMap:end")
    return tokens

  def object(self, _=False):
    return {o[0]: o[1] for o in [f.object(_) for f in self.fields]}

class Field(ASTNode):
  name = None
  value = None

  def __str__(self):
    return "Field ({}) with value:\n  {}".format(self.name, indent(self.value))

  def parse(self, tokens):
    debug.debug("Parsing", "Field:start")
    debug.indent("Parsing")
    self.name = Literal(self)
    tokens = self.name.parse(tokens)
    vtype = lookup_value_type(self.parent.parent.typ.object(), self.name.value)
    debug.debug("Parsing", "Field:type={} ({}.{})".format(
      vtype,
      self.parent.parent.typ.name,
      self.name.value,
    ))
    self.value = vtype(self)
    tokens = self.value.parse(tokens)
    debug.exdent("Parsing")
    debug.debug("Parsing", "Field:end")
    return tokens

  def object(self, _=False):
    return (self.name.object(_), self.value.object(_))

class ObjectMap(ASTNode):
  objects = set()

  def __str__(self):
    return "ObjectMap with contents:\n  {}".format(
      indent(', '.join(str(o) for o in self.objects))
    )

  def parse(self, tokens):
    debug.debug("Parsing", "ObjectMap:start")
    debug.indent("Parsing")
    items, tokens = self.parse_list(tokens, Object, False)
    self.objects = set(items)
    debug.exdent("Parsing")
    debug.debug("Parsing", "ObjectMap:end")
    return tokens

  def object(self, _=False):
    return {o.id: o for o in [obj.object(_) for obj in self.objects]}

class Object(ASTNode):
  typ = None
  id = None
  fields = None
  original = None

  def __str__(self):
    return "Object of type {} with id {} and fields:\n  {}".format(
      self.typ,
      self.id,
      indent(self.fields),
    )

  def parse(self, tokens):
    debug.debug("Parsing", "Object:start")
    debug.indent("Parsing")
    self.typ = Type(self)
    tokens = self.typ.parse(tokens)
    self.id = Literal(self)
    tokens = self.id.parse(tokens)
    self.fields = FieldMap(self)
    tokens = self.fields.parse(tokens)
    debug.exdent("Parsing")
    debug.debug("Parsing", "Object:end")
    return tokens

  def object(self, rebuild=False):
    if rebuild:
      if not self.original:
        raise ParseError("Can't rebuild unbuilt object.", self, ["<unknown>"])
      if isinstance(self.original, text.Anything):
        fields = self.fields.object(rebuild)
        for f in fields:
          if f == "id":
            continue
          setattr(self.original, f, fields[f])
        return self.original
      elif isinstance(self.original, actions.Action):
        debug.debug("Parsing", "Action rebuild request ignored.")
        return self.original
      else:
        raise ParseError(
          "We don't have the technology to rebuild that.",
          self,
          ["<unknown>"]
        )
    else:
      typ = self.typ.object(rebuild)
      id = self.id.object(rebuild)
      fields = self.fields.object(rebuild)
      debug.debug(
        "Construction",
        "Constructing object of type {} with id '{}' from fields:\n  {}".format(
          typ,
          id,
          indent(fields),
        )
      )
      obj = typ(id=id, **fields)
      self.original = obj
      return obj

class BindingMap(ASTNode):
  bindings = []

  def __str__(self):
    return "BindingMap with contents:\n  {}".format(indent(self.bindings))

  def parse(self, tokens):
    debug.debug("Parsing", "BindingMap:start")
    debug.indent("Parsing")
    self.bindings, tokens = self.parse_list(tokens, Binding, False)
    debug.exdent("Parsing")
    debug.debug("Parsing", "BindingMap:end")
    return tokens

  def object(self, _=False):
    result = collections.OrderedDict()
    for b in self.bindings:
      bo = b.object(_)
      key = '{type}/{name}'.format(type=bo[0], name=bo[1])
      result[key] = bo[2]
    return result

class Binding(ASTNode):
  typ = None
  name = None
  condlist = None

  def __str__(self):
    return "Binding of type {} with name {} and conditions:\n  {}".format(
      self.typ,
      self.name,
      indent(self.condlist),
    )

  def parse(self, tokens):
    debug.debug("Parsing", "Binding:start")
    debug.indent("Parsing")
    self.typ = Type(self)
    tokens = self.typ.parse(tokens)
    self.name = Literal(self)
    tokens = self.name.parse(tokens)
    self.condlist = List.of(Literal)(self)
    debug.exdent("Parsing")
    debug.debug("Parsing", "Binding:end")
    return self.condlist.parse(tokens)

  def object(self, _=False):
    return (
      self.typ.object(_).__name__,
      self.name.object(_),
      self.condlist.object(_)
    )

########################
# Generic node classes #
########################

class List(ASTNode):
  typ = Literal
  classname = "List"
  items = []

  def __str__(self):
    return "{} with contents:\n  {}".format(
      self.classname,
      indent(self.items),
    )

  @staticmethod
  def of(type):
    class TypedList(List):
      nonlocal type
      typ = type
      classname = "List<{}>".format(type)
    return TypedList

  def parse(self, tokens):
    debug.debug("Parsing", "{}:start".format(self.classname))
    debug.indent("Parsing")
    self.items, tokens = self.parse_list(tokens, self.typ)
    debug.exdent("Parsing")
    debug.debug("Parsing", "{}:end".format(self.classname))
    return tokens

  def object(self, _=False):
    return [it.object(_) for it in self.items]

class Set(List):
  classname = "Set"

  @staticmethod
  def of(type):
    class TypedSet(Set):
      nonlocal type
      typ = type
      classname = "Set<{}>".format(type)
    return TypedSet

  def object(self, _=False):
    return set(it.object(_) for it in self.items)

class Map(ASTNode):
  keytyp = Literal
  valtyp = Literal
  classname = "Map"
  mapping = collections.OrderedDict()

  def __str__(self):
    return "{} with contents:\n  {}".format(
      self.classname,
      indent('\n'.join(str(k) + ' : ' + str(v)
                       for k, v in self.mapping.items())),
    )

  @staticmethod
  def of(keytype, valtype):
    class TypedMap(Map):
      nonlocal keytype, valtype
      keytyp = keytype
      valtyp = valtype
      classname = "Map<{}, {}>".format(keytype, valtype)
    return TypedMap

  def parse(self, tokens):
    debug.debug("Parsing", "{}:start".format(self.classname))
    debug.indent("Parsing")
    self.mapping = collections.OrderedDict()
    # TODO: Implement a pair ASTNode so that this can be a call to parse_list
    if tuple(tokens[:2]) == (':', '\n'):
      tokens = tokens[2:]
      while tokens[0] in ('', '\n'):
        tokens = tokens[1:]
      while tokens[0] != 'end':
        key = self.keytyp(self)
        tokens = key.parse(tokens)
        value = self.valtyp(self)
        tokens = value.parse(tokens)
        self.mapping[key] = value
        while tokens[0] in ('', '\n'):
          tokens = tokens[1:]
    else:
      raise ParseError(
        "Maps must use the ': ... end' listing style.",
        self,
        tokens
      )
    debug.exdent("Parsing")
    debug.debug("Parsing", "{}:end".format(self.classname))
    if tuple(tokens[:2]) == ('end', '\n'):
      return tokens[2:]
    else:
      return tokens[1:]

  def object(self, _=False):
    return {key.object(_): value.object(_)
            for key, value in self.mapping.items()}

##########################
# Reference node classes #
##########################

class IDReference(ASTNode):
  id = None
  original = None

  def __str__(self):
    return "Reference to {}".format(self.id)

  def parse(self, tokens):
    debug.debug("Parsing", "IDReference:start")
    debug.indent("Parsing")
    self.id = Literal(self)
    tokens = self.id.parse(tokens)
    debug.exdent("Parsing")
    debug.debug("Parsing", "IDReference:end")
    return tokens

  def object(self, tools=False):
    if tools:
      if not self.original:
        raise ParseError(
          "Can't resolve an unbuilt reference.",
          self,
          ["<unknown>"]
        )
      return self.original.resolve(*tools)
    else:
      self.original = Reference(self.id.object(tools))
      return self.original

class IDGenReference(ASTNode):
  id = None
  typ = None
  classname = "IDGenReference"

  def __str__(self):
    return "Generative Reference of type {} to {}".format(
      self.typ,
      self.id,
    )

  @staticmethod
  def to(type):
    class TypedIDGR(IDGenReference):
      nonlocal type
      typ = type
      classname = "IDGenReference<{}>".format(type)
    return TypedIDGR

  def parse(self, tokens):
    debug.debug("Parsing", "{}:start".format(self.classname))
    debug.indent("Parsing")
    self.id = Literal(self)
    tokens = self.id.parse(tokens)
    debug.exdent("Parsing")
    debug.debug("Parsing", "{}:end".format(self.classname))
    return tokens

  def object(self, tools):
    if tools:
      if not self.original:
        raise ParseError(
          "Can't resolve an unbuilt reference.",
          self,
          ["<unknown>"]
        )
      return self.original.resolve(*tools)
    else:
      self.original = GenRef(self.id.object(tools), self.typ)
      return self.original

# A mapping from parent object type to mappings from field names to field value
# AST Node types (inheritance is in effect):
FIELD_VALUE_TYPES = {
  text.Anything: {
    "properties": Set.of(Literal),
  },
  text.TEXT: {
    "starting_place": IDReference,
    "starting_time": IDReference,
    "directions": ObjectMap,
    "places": ObjectMap,
    "times": ObjectMap,
    "things": ObjectMap,
    "events": ObjectMap,
    "conditions": ObjectMap,
    "actions": ObjectMap,
  },
  text.Direction: {
    "abbreviations": Set.of(Literal),
    "opposite": IDGenReference.to(text.Direction),
  },
  text.Place: {
    "order": Number,
    "connections": Map.of(IDGenReference.to(text.Direction), IDReference),
  },
  text.Time: {
    "order": Number,
  },
  text.Thing: {
    "plural": ImpliedTrue,
    "starting_place": IDReference,
  },
  text.Event: {
    "conditions": List.of(Literal),
    "bindings": BindingMap,
    "order": Number,
  },
  text.Condition: {
    "bindings": BindingMap,
    "requires": List.of(Literal),
  },
}

def assemble_AST(data):
  '''
  Assembles an Abstract Syntax Tree from given data.
  '''
  tokens, comments = tokenize(data)
  AST = Object(None)
  AST.typ = Type(AST)
  AST.typ.name = "TEXT"
  AST.id = Literal(AST)
  AST.id.value = "the text"
  AST.fields = FieldMap(AST)
  remainder = AST.fields.parse([':', '\n'] + tokens + ['end'])
  if any(t not in ('', '\n') for t in remainder):
    raise ParseError("Leftover tokens after parse.", AST, remainder)
  return AST

def tokenize(data):
  '''
  Splits the given data into tokens, separating normal tokens from comment
  tokens and handling {}-quoting and line-terminal ':' characters.
  '''
  tokens = ['']
  comments = [''] # TODO: Improve this?
  state = None
  for c in data:
    if state == None:
      if c == '#':
        state = 'comment'
        if tokens[-1]:
          tokens.append('')
      elif c == '{':
        state = 'block'
      elif c in (' ','\t', '\n'):
        if c == '\n':
          tokens.append('\n')
        if tokens[-1]:
          tokens.append('')
      elif c == ':':
        tokens[-1] += c
        state = "colonmightbeterminal"
      else:
        tokens[-1] += c
    elif state == "colonmightbeterminal":
      if c == '\n':
        tokens[-1] = tokens[-1][:-1]
        tokens.append(':')
        tokens.append('\n')
        tokens.append('')
        state = None
      elif c == ':':
        tokens[-1] += c
      else:
        tokens[-1] += c
        state = None
    elif state == 'comment':
      if c == '\n':
        comments.append('')
        state = None
      elif c == '#' and comments[-1] == '':
        state = 'supercomment'
      else:
        comments[-1] += c
    elif state == 'supercomment':
      if c == '#':
        state = 'supercommentmightend'
        comments[-1] += c
      else:
        comments[-1] += c
    elif state == 'supercommentmightend':
      if c == '#':
        state = None
        comments[-1] = comments[-1][:-1]
        comments.append('')
      else:
        state = 'supercomment'
        comments[-1] += c
    elif state == 'block':
      if c == '{':
        state = 'expectingopenbrace'
        tokens[-1] += c
      elif c == '}':
        state = 'blockmightend'
        tokens[-1] += c
      else:
        tokens[-1] += c
    elif state == 'expectingopenbrace':
      if c == '{':
        state = 'block'
      else:
        tokens[-1] += c
        debug.debug(
          "Warning: Data contained lonely '{' within a block ({}).".format(
            tokens[-1]
          )
        )
        state = 'block'
    elif state == 'blockmightend':
      if c == '}':
        state = 'block'
      else:
        tokens[-1] = tokens[-1][:-1]
        state = None
        if c == '#':
          state = 'comment'
          tokens.append('')
        elif c == '{':
          state = 'block'
        elif c in (' ','\t', '\n'):
          if c == '\n':
            tokens.append('\n')
          tokens.append('')
        elif c == ':':
          tokens[-1] += c
          state = "colonmightbeterminal"
        else:
          tokens[-1] += c
  if state != None:
    debug.debug("Warning: tokenization ended in state '{}'".format(state))
  if tokens[-1] == '':
    tokens = tokens[:-1]
  if comments[-1] == '':
    comments = comments[:-1]
  return tokens, comments
