#!/usr/bin/env python3

import optparse, sys

try:
  import readline
except ImportError:
  readline = False

import text.debug, text.parser

def OPTIONS(parser):
  """
  Defines the options that textension accepts.
  """
  parser.add_option(
    "-d",
    "--debug",
    action="store_true",
    help="Turn on (quite verbose) debugging.",
    dest="DEBUG",
  )
  parser.add_option(
    "-L",
    "--log-file",
    default="log",
    dest="logfile",
    help="Set the log file (default is 'log').",
    metavar="LOGFILE",
  )
  parser.add_option(
    "-s",
    "--show-stats",
    default=False,
    action="store_true",
    help="Print the stats line in the introduction.",
    dest="stats",
  )
  parser.add_option(
    "-l",
    "--load",
    dest="tracefile",
    help="Start the session by executing actions contained in the given trace "
         "file.",
    metavar="TRACE",
  )
  parser.add_option(
    "-q",
    "--quit",
    action="store_true",
    dest="quit",
    help="Quit immediately after running any trace data."
  )

def print_description(TEXT):
  print
  print(TEXT.describe(), end="")

def repl(TEXT):
  """
  Runs the Read, Evaluate, Print Loop to execute the given Temporal EXTension.
  """
  if readline:
    readline.clear_history()
    readline.set_history_length(-1)
    if options.tracefile:
      readline.read_history_file(options.tracefile)
  TEXT.init()
  done = False
  if options.tracefile:
    try:
      with open(options.tracefile, 'rU') as fin:
        for line in fin:
          if line[-1] == '\n':
            line = line[:-1]
          print('-'*80)
          print_description(TEXT)
          print('-'*80)
          print(">" + line)
          print('-'*80)
          TEXT.handle_action(line)
    except:
      print("Aborting trace load because of error:")
      sys.excepthook(*sys.exc_info())
      done = True
  if options.quit:
    print('-'*80)
    print_description(TEXT)
    print('-'*80)
    done = True
  print('\n'*60)
  print('-'*80)
  print(TEXT.introduce(stats=options.stats))
  while not done:
    try:
      print('-'*80)
      print_description(TEXT)
      print('-'*80)
      try:
        action = input(">")
        print('-'*80)
        print('\n'*60)
        print('-'*80)
      except EOFError:
        action = ""
        done = True
      if action in ('q', 'quit'):
        if readline:
          readline.remove_history_item(
            readline.get_current_history_length() - 1
          )
        done = True
      else:
        TEXT.handle_action(action)
    except KeyboardInterrupt:
      if text.debug.domainactive("ForcedExit"):
        sys.excepthook(*sys.exc_info())
      done = True
    except:
      print("Aborting because of error:")
      sys.excepthook(*sys.exc_info())
      done = True
  print()
  if readline:
    readline.write_history_file('trace')

def run(filename):
  '''
  Reads data from the given file and runs the resulting TEXT.
  '''
  with open(filename, 'rU') as fin:
    data = fin.read()
  TEXT = text.parser.parse_TEXT(data)
  repl(TEXT)

if __name__ == "__main__":
  global options
  optparser = optparse.OptionParser(
    usage="""\
%prog [options] TEXTFILE

  Runs the Temporal EXTension specified within TEXTFILE.\
""",
    version="%prog 0.3",
  )
  OPTIONS(optparser)
  (options, args) = optparser.parse_args()
  text.debug.set_log(options.logfile)
  if options.DEBUG:
    text.debug.activate()
  if not args:
    optparser.print_help()
  else:
    run(args[0])
#    import cProfile
#    cProfile.run("run(args[0])")
